from numba_stats import crystalball_ex as cb
import numpy as np
import pytest
from scipy import stats as sc
from numpy.testing import assert_allclose
from scipy.integrate import quad


@pytest.mark.parametrize("beta", (0.1, 2, 3))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_pdf_left(beta, m):
    scale = 1.5
    x = np.linspace(-10, 0, 10)
    got = cb.pdf(x, beta, m, scale, 2 * beta, 2 * m, 2 * scale, 0)

    max1 = cb.pdf(0, beta, m, scale, 2 * beta, 2 * m, 2 * scale, 0)
    max2 = sc.crystalball.pdf(0, beta, m, 0, scale)

    expected = sc.crystalball.pdf(x, beta, m, 0, scale) * max1 / max2
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (0.1, 2, 3))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_pdf_right(beta, m):
    scale = 1.5
    x = np.linspace(-10, 0, 10)
    got = cb.pdf(-x, 2 * beta, 2 * m, 2 * scale, beta, m, scale, 0)

    max1 = cb.pdf(0, beta, m, scale, 2 * beta, 2 * m, 2 * scale, 0)
    max2 = sc.crystalball.pdf(0, beta, m, 0, scale)

    expected = sc.crystalball.pdf(x, beta, m, 0, scale) * max1 / max2
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (0.1, 2, 3))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_pdf_integral(beta, m):
    scale = 1.5
    got = quad(
        lambda x: cb.pdf(x, 2 * beta, 2 * m, 2 * scale, beta, m, scale, 0),
        -np.inf,
        np.inf,
    )[0]
    assert_allclose(got, 1)


@pytest.mark.parametrize("beta", (0.1, 2, 3))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_cdf(beta, m):
    scale = 1.5
    x = np.linspace(-10, 10, 10)
    got = cb.cdf(x, beta, m, scale, 2 * beta, 2 * m, 2 * scale, 0)
    expected = [
        quad(
            lambda x: cb.pdf(x, beta, m, scale, 2 * beta, 2 * m, 2 * scale, 0),
            -np.inf,
            xi,
        )[0]
        for xi in x
    ]
    assert_allclose(got, expected)



import scipy.stats as sc
import numpy as np
from numba_stats import t
import pytest


@pytest.mark.parametrize("df", (1, 1.5, 2, 3, 4, 5.5, 10))
def test_pdf(df):
    x = np.linspace(-5, 5, 10)
    got = t.pdf(x, df, 2, 3)
    expected = sc.t.pdf(x, df, 2, 3)  # supports real-valued df
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("df", (1, 1.5, 3, 5.5, 10))
def test_cdf(df):
    x = np.linspace(-5, 5, 10)
    got = t.cdf(x, df, 2, 3)
    expected = sc.t.cdf(x, df, 2, 3)  # supports real-valued df
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("df", (1, 1.5, 3, 5.5, 10))
def test_ppf(df):
    x = np.linspace(0, 1, 10)
    got = t.ppf(x, df, 2, 3)
    expected = sc.t.ppf(x, df, 2, 3)  # supports real-valued df
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("df", (1, 1.5, 3, 5.5, 10))
def test_rvs(df):
    args = df, 2, 3
    x = t.rvs(*args, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: t.cdf(x, *args))
    assert r.pvalue > 0.01



from numba_stats import bernstein
from scipy.interpolate import BPoly
import pytest
import numpy as np
from scipy.integrate import quad
import numba as nb
from numpy.testing import assert_allclose


def scipy_density(x, beta, xmin, xmax):
    return BPoly(np.array(beta)[:, np.newaxis], [xmin, xmax])(x)


@pytest.mark.parametrize(
    "beta", [[1.0], [1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 2.0, 3.0], [1.0, 3.0, 2.0]]
)
def test_density(beta):
    x = np.linspace(1, 3, 10000)
    got = bernstein.density(x, beta, x[0], x[-1])
    expected = scipy_density(x, beta, x[0], x[-1])
    assert_allclose(got, expected)

    got = bernstein.density(0.5, beta, 0, 1)
    expected = bernstein.density([0.5], beta, 0, 1)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", [[1], [1, 1], [1, 1, 1]])
def test_integral(beta):
    xrange = 1.5, 3.4
    got = bernstein.integral(xrange[1], beta, *xrange)
    expected = np.diff(xrange)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", [[1], [1, 1], [1, 1, 1], [1, 2, 3], [1, 3, 2]])
def test_integral_2(beta):
    x = np.linspace(1, 2.5)

    got = bernstein.integral(x, beta, x[0], x[-1])
    expected = [
        quad(lambda y: bernstein.density(y, beta, x[0], x[-1]), x[0], xi)[0] for xi in x
    ]
    assert_allclose(got, expected)


@pytest.mark.filterwarnings("error")
@pytest.mark.parametrize("parallel", (False, True))
@pytest.mark.parametrize("fn", [bernstein.density, bernstein.integral])
def test_numba(fn, parallel):
    x = np.linspace(0.5, 0.6, 10000)
    beta = np.array([1.0, 2.0, 3.0])
    xmin = 1.0
    xmax = 2.5

    @nb.njit(parallel=parallel, fastmath=True)
    def f():
        return fn(x, beta, xmin, xmax)

    assert_allclose(f(), fn(x, beta, xmin, xmax))


def test_deprecation():
    with pytest.warns(np.VisibleDeprecationWarning):
        got = bernstein.scaled_pdf(1, [1, 2], 0, 1)
    assert_allclose(got, bernstein.density(1, [1, 2], 0, 1))

    with pytest.warns(np.VisibleDeprecationWarning):
        got = bernstein.scaled_cdf(1, [1, 2], 0, 1)
    assert_allclose(got, bernstein.integral(1, [1, 2], 0, 1))



"""
Cruijff distribution.

See For example: https://arxiv.org/abs/1005.4087
"""

from ._util import _jit, _generate_wrappers, _prange
import numpy as np

_doc_par = """
beta_left: float
    Left tail acceleration parameter.
beta_right: float
    Right tail acceleration parameter.
loc : float
    Location of the maximum of the distribution.
scale_left : float
    Left width parameter.
scale_right: float
    Right width parameter.
"""


@_jit(5)
def _density(x, beta_left, beta_right, loc, scale_left, scale_right):
    r = np.empty_like(x)
    for i in _prange(len(x)):
        if x[i] < loc:
            scale = scale_left
            beta = beta_left
        else:
            scale = scale_right
            beta = beta_right
        z = (x[i] - loc) / scale
        r[i] = -0.5 * z**2 / (1 + beta * z**2)
    return np.exp(r)


_generate_wrappers(globals())



import numba as nb
import numpy as np
from numba.types import Array
from numba.core.errors import TypingError
from numba.extending import overload
from numba import prange as _prange  # noqa
import os

_Floats = (nb.float32, nb.float64)


def _readonly_carray(T):
    return Array(T, 1, "A", readonly=True)


def _jit(arg, cache=True):
    """
    Wrapper for numba.njit to reduce boilerplate code.

    We want to build jitted functions with explicit signatures to restrict the argument
    types which are used in the implemnetation to float32 or float64. We also want to
    pass specific options consistently: error_model='numpy' and inline='always'. The
    latter is important to profit from auto-parallelization of surrounding code.

    Parameters
    ----------
    arg : int
        Number of arguments. If negative, all arguments of this function are scalars
        and -arg is the number of arguments. If positive, the first argument is
        an array, the others are scalars and arg is the number of scalar arguments.
    """
    if isinstance(arg, list):
        signatures = arg
    else:
        signatures = []
        for T in (nb.float32, nb.float64):
            if arg < 0:
                sig = T(*([T] * -arg))
            else:
                sig = T[:](_readonly_carray(T), *[T for _ in range(arg)])
            signatures.append(sig)
    return nb.njit(signatures, cache=cache, inline="always", error_model="numpy")


def _rvs_jit(arg, cache=True):
    signatures = []
    T = nb.float64  # nb.float32 cannot be supported
    # extra args at the end are for size and random_state
    sig = T[:](*[T for _ in range(arg)], nb.uint64, nb.optional(nb.uint64))
    signatures.append(sig)
    return nb.njit(signatures, cache=cache, inline="always", error_model="numpy")


@nb.njit(cache=True)
def _seed(seed):
    if seed is None:
        with nb.objmode(seed="optional(uint8)"):
            seed = np.frombuffer(os.urandom(8), dtype=np.uint64)[0]
    np.random.seed(seed)


def _wrap(fn):
    def outer(first, *rest):
        shape = np.shape(first)
        first = np.array(first).flatten()
        if first.dtype.kind != "f":
            first = first.astype(float)
        return fn(first, *rest).reshape(shape)

    return outer


@_jit(2)
def _trans(x, loc, scale):
    inv_scale = type(scale)(1) / scale
    return (x - loc) * inv_scale


def _type_check(first, *rest):
    if not (isinstance(first, Array) and first.dtype in _Floats):
        raise TypingError("first argument must be an array of floating point type")

    T = type(first.dtype)
    for i, tp in enumerate(rest):
        if not isinstance(tp, T):
            raise TypingError(f"argument {i+1} must be of type {tp}")


def _generate_wrappers(d):
    import inspect

    if "_wrap" not in d:
        d["_wrap"] = _wrap
    if "_type_check" not in d:
        d["_type_check"] = _type_check
    d["_overload"] = overload

    doc_par = d["_doc_par"].strip() if "_doc_par" in d else None

    for fname in (
        "pdf",
        "pmf",
        "logpdf",
        "logpmf",
        "cdf",
        "ppf",
        "density",
        "integral",
        "rvs",
    ):
        impl = f"_{fname}"
        if impl not in d:
            continue
        fn = d[impl]
        args = inspect.signature(fn).parameters
        args = ", ".join([f"{x}" for x in args])
        doc_title = {
            "density": "Return density.",
            "integral": "Return integrated density.",
            "logpdf": "Return log of probability density.",
            "logpmf": "Return log of probability mass.",
            "pmf": "Return probability mass.",
            "pdf": "Return probability density.",
            "cdf": "Return cumulative probability.",
            "ppf": "Return quantile for given probability.",
            "rvs": "Return random samples from distribution.",
        }.get(fname, None)
        if fname == "ppf":
            before_par = """\
x: ArrayLike
    Probability. Must be between 0 and 1.
"""
        elif fname == "rvs":
            before_par = ""
        else:
            before_par = """\
x: ArrayLike
    Random variate.
"""
        if fname == "rvs":
            after_par = """
size : int
    Number of random variates.
random_state : int or None
    Seed of the random number generator. Default is None, which uses a random seed."""
        else:
            after_par = ""

        if fname == "rvs":
            code = f"""
def {fname}({args}):
    return {impl}({args})

@_overload({fname}, inline="always")
def _ol_{fname}({args}):
    return {impl}
"""
        else:
            code = f"""
def {fname}({args}):
    return _wrap({impl})({args})

@_overload({fname}, inline="always")
def _ol_{fname}({args}):
    _type_check({args})
    return {impl}.__wrapped__
"""

        if doc_par is None:
            code += f"""
{fname}.__doc__ = {impl}.__doc__
"""
        else:
            assert doc_title is not None
            code += f"""
{fname}.__doc__ = \"\"\"
{doc_title}

Parameters
----------
{before_par}{doc_par}{after_par}

Returns
-------
Array-like
    Function evaluated at the x points.
\"\"\"
"""
        exec(code, d)



"""
Poisson distribution.

See Also
--------
scipy.stats.poisson: Scipy equivalent.
"""

import numpy as np
from ._special import gammaincc as _gammaincc
from math import lgamma as _lgamma
from ._util import _jit, _generate_wrappers, _prange

_doc_par = """
mu : float
    Expected value.
"""


@_jit(1)
def _logpmf(k, mu):
    T = type(mu)
    r = np.empty(len(k), T)
    for i in _prange(len(r)):
        if mu == 0:
            r[i] = 0.0 if k[i] == 0 else -np.inf
        else:
            r[i] = k[i] * np.log(mu) - _lgamma(k[i] + T(1)) - mu
    return r


@_jit(1)
def _pmf(k, mu):
    return np.exp(_logpmf(k, mu))


@_jit(1, cache=False)
def _cdf(k, mu):
    T = type(mu)
    r = np.empty(len(k), T)
    for i in _prange(len(r)):
        r[i] = _gammaincc(k[i] + T(1), mu)
    return r


_generate_wrappers(globals())



"""
Q-Gaussian distribution.

A generalisation (q-analog) of the normal distribution based on Tsallis entropy. It
can be used an alternative model for the normal distribution to check for model bias.

It is equivalent to Student's t distribution and can be computed from the latter via
a change of variables, which is exploited in this implementation.

https://en.wikipedia.org/wiki/Q-Gaussian_distribution
"""

import numpy as np
import numba as nb
from . import norm as _norm, t as _t
from ._util import _jit, _generate_wrappers, _rvs_jit

_doc_par = """
q : float
    Shape parameter between 1 and 3. For q = 1, the qgaussian is a normal distribution,
    for q == 3 it is a cauchy distribution.
loc : float
    Expectation value.
scale : float
    Width parameter.
"""


@nb.njit
def _df_sigma(q, sigma):
    # https://en.wikipedia.org/wiki/Q-Gaussian_distribution
    # relation to Student's t-distribution

    # 1/(2 sigma^2) = 1 / (3 - q)
    # 2 sigma^2 = 3 - q
    # sigma = sqrt((3 - q)/2)
    T = type(q)
    df = (T(3) - q) / (q - T(1))
    sigma /= np.sqrt(T(0.5) * (T(3) - q))

    return df, sigma


@_jit(3)
def _logpdf(x, q, mu, sigma):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm._logpdf(x, mu, sigma)

    df, sigma = _df_sigma(q, sigma)

    return _t._logpdf(x, df, mu, sigma)


@_jit(3)
def _pdf(x, q, mu, sigma):
    return np.exp(_logpdf(x, q, mu, sigma))


@_jit(3, cache=False)
def _cdf(x, q, mu, sigma):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm._cdf(x, mu, sigma)

    df, sigma = _df_sigma(q, sigma)

    return _t._cdf(x, df, mu, sigma)


@_jit(3, cache=False)
def _ppf(p, q, mu, sigma):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm._ppf(p, mu, sigma)

    df, sigma = _df_sigma(q, sigma)

    return _t._ppf(p, df, mu, sigma)


@_rvs_jit(3, cache=False)
def _rvs(q, mu, sigma, size, random_state):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm._rvs(mu, sigma, size, random_state)

    df, sigma = _df_sigma(q, sigma)

    return _t._rvs(df, mu, sigma, size, random_state)


_generate_wrappers(globals())



import scipy.stats
import numpy as np
import numba as nb
import pytest
import importlib

N = (10, 100, 1000, 10_000, 100_000)

KIND = (
    "norm.pdf",
    "norm.cdf",
    "norm.ppf",
    "truncnorm.pdf",
    "truncnorm.cdf",
    "truncnorm.ppf",
    "expon.pdf",
    "expon.cdf",
    "expon.ppf",
    "truncexpon.pdf",
    "truncexpon.cdf",
    "truncexpon.ppf",
    "t.pdf",
    "t.cdf",
    "t.ppf",
    "uniform.pdf",
    "uniform.cdf",
    "uniform.ppf",
)

ARGS = {
    "norm": (0.1, 1.2),
    "truncnorm": (-1.0, 1.0, 0.0, 1.0),
    "expon": (0.0, 1.4),
    "truncexpon": (0.0, 1.0, 0.0, 1.0),
    "t": (5.0, 0.0, 1.0),
    "uniform": (0.0, 1.0),
}


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("lib", ("scipy", "ours", "ours:parallel,fastmath"))
@pytest.mark.parametrize("kind", KIND)
def test_speed(benchmark, lib, kind, n):
    dist, meth = kind.split(".")

    if kind == "t.ppf" and lib == "ours:parallel,fastmath":
        pytest.skip("broken in parallel mode")

    if meth == "ppf":
        x = np.linspace(0, 1, n)
    else:
        x = np.linspace(-5 if dist != "expon" else 0, 5, n)

    rng = np.random.default_rng(1)
    rng.shuffle(x)

    args = ARGS[dist]
    if dist == "truncexpon" and lib == "scipy":
        args = args[1:]

    if dist == "voigt" and lib == "scipy":
        from scipy.special import voigt_profile as method
    else:
        if lib == "scipy":
            d = getattr(scipy.stats, dist)
        else:
            d = importlib.import_module(f"numba_stats.{dist}")
        method = getattr(d, meth)
        if lib == "ours:parallel,fastmath":
            orig = method

            if len(args) == 2:

                @nb.njit(parallel=True, fastmath=True)
                def method(x, a, b):
                    return orig(x, a, b)

            elif len(args) == 3:

                @nb.njit(parallel=True, fastmath=True)
                def method(x, a, b, c):
                    return orig(x, a, b, c)

            elif len(args) == 4:

                @nb.njit(parallel=True, fastmath=True)
                def method(x, a, b, c, d):
                    return orig(x, a, b, c, d)

            else:
                assert False

    # warm-up JIT
    method(x, *args)
    benchmark(method, x, *args)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("lib", ("scipy", "ours", "ours:parallel,fastmath"))
def test_speed_voigt_pdf(benchmark, lib, n):
    x = np.linspace(-5, 5, n)
    rng = np.random.default_rng(1)
    rng.shuffle(x)

    sigma = 1.2
    gamma = 0.5

    if lib == "scipy":
        from scipy.special import voigt_profile

        def method(x, gamma, loc, sigma):
            return voigt_profile(x - loc, sigma, gamma)

    else:
        from numba_stats import voigt

        if lib == "ours:parallel,fastmath":

            @nb.njit(parallel=True, fastmath=True)
            def method(x, gamma, loc, sigma):
                return voigt.pdf(x, gamma, loc, sigma)

        else:
            method = voigt.pdf

    # warm-up JIT
    method(x, gamma, 0.0, sigma)
    benchmark(method, x, gamma, 0.0, sigma)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("lib", ("scipy", "ours", "ours:parallel,fastmath"))
def test_speed_bernstein_density(benchmark, lib, n):
    from numba_stats import bernstein
    from scipy.interpolate import BPoly

    x = np.linspace(0, 1, n)
    rng = np.random.default_rng(1)
    rng.shuffle(x)

    beta = np.arange(1, 4, dtype=float)
    xmin = np.min(x)
    xmax = np.max(x)

    if lib == "scipy":

        def method(x, beta, xmin, xmax):
            return BPoly(np.array(beta)[:, np.newaxis], [xmin, xmax])(x)

    elif lib == "ours:parallel,fastmath":
        method = bernstein.density

        if lib == "ours:parallel,fastmath":
            method = nb.njit(parallel=True, fastmath=True)(method)

    # warm-up JIT
    method(x, beta, xmin, xmax)
    benchmark(method, x, beta, xmin, xmax)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("lib", ("scipy", "ours", "ours:parallel,fastmath"))
def test_speed_truncexpon_pdf_plus_norm_pdf(benchmark, lib, n):
    x = np.linspace(0, 1, n)
    rng = np.random.default_rng(1)
    rng.shuffle(x)

    xmin = np.min(x)
    xmax = np.max(x)

    if lib == "scipy":
        from scipy.stats import norm, truncexpon

        def method(x, z, mu, sigma, slope):
            p1 = truncexpon.pdf(x, xmax, xmin, slope)
            p2 = norm.pdf(x, mu, sigma)
            return (1 - z) * p1 + z * p2

    else:
        from numba_stats import norm, truncexpon

        def method(x, z, mu, sigma, slope):
            p1 = truncexpon.pdf(x, xmin, xmax, 0.0, slope)
            p2 = norm.pdf(x, mu, sigma)
            return (1 - z) * p1 + z * p2

        if lib == "ours:parallel,fastmath":
            method = nb.njit(parallel=True, fastmath=True)(method)

    # warm-up JIT
    args = 0.5, 0.5, 0.1, 1.0
    method(x, *args)
    benchmark(method, x, *args)



"""
Generalised Crystal Ball distribution.

The generalised Crystal Ball distribution replaces the lower and upper tail of
an asymmetric normal distribution with power-law tails. Furthermore, the scale
is allowed to vary between the left and the right side of the peak. There is no
discontinuity at the maximum or elsewhere.
"""

from .crystalball import _powerlaw_integral, _normal_integral, _log_density
from ._util import _jit, _generate_wrappers, _prange
import numpy as np

_doc_par = """
beta_left : float
    Distance from the mode in units of standard deviations where the Crystal Ball
    turns from a gaussian into a power law on the left side.
m_left : float
    Absolute value of the slope of the left powerlaw tail. Must be large than 1.
scale_left : float
    Standard deviation of the left side of the mode.
beta_right : float
    Distance from the mode in units of standard deviations where the Crystal Ball
    turns from a gaussian into a power law on the right side.
m_right : float
    Absolute value of the slope of the right powerlaw tail. Must be large than 1.
scale_right : float
    Standard deviation of the right side of the mode.
loc : float
    Location of the mode of the distribution.
"""


@_jit(-3)
def _norm_half(beta, m, scale):
    T = type(beta)
    return (_powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, T(0))) * scale


@_jit(7)
def _logpdf(x, beta_left, m_left, scale_left, beta_right, m_right, scale_right, loc):
    norm = _norm_half(beta_left, m_left, scale_left) + _norm_half(
        beta_right, m_right, scale_right
    )
    c = np.log(norm)
    r = np.empty_like(x)
    for i in _prange(len(r)):
        if x[i] < loc:
            beta = beta_left
            m = m_left
            z = (x[i] - loc) / scale_left
        else:
            beta = beta_right
            m = m_right
            z = (loc - x[i]) / scale_right
        r[i] = _log_density(z, beta, m) - c
    return r


@_jit(7)
def _pdf(x, beta_left, m_left, scale_left, beta_right, m_right, scale_right, loc):
    return np.exp(
        _logpdf(
            x,
            beta_left,
            m_left,
            scale_left,
            beta_right,
            m_right,
            scale_right,
            loc,
        )
    )


@_jit(7)
def _cdf(x, beta_left, m_left, scale_left, beta_right, m_right, scale_right, loc):
    T = type(beta_left)
    norm = _norm_half(beta_left, m_left, scale_left) + _norm_half(
        beta_right, m_right, scale_right
    )
    r = np.empty_like(x)
    for i in _prange(len(x)):
        scale = T(1) / (scale_left if x[i] < loc else scale_right)
        z = (x[i] - loc) * scale
        if z < -beta_left:
            r[i] = _powerlaw_integral(z, beta_left, m_left) * scale_left / norm
        elif z < 0:
            r[i] = (
                (
                    _powerlaw_integral(-beta_left, beta_left, m_left)
                    + _normal_integral(-beta_left, z)
                )
                * scale_left
                / norm
            )
        elif z < beta_right:
            r[i] = (
                (
                    _powerlaw_integral(-beta_left, beta_left, m_left)
                    + _normal_integral(-beta_left, T(0))
                )
                * scale_left
                + _normal_integral(T(0), z) * scale_right
            ) / norm
        else:
            r[i] = (
                (
                    _powerlaw_integral(-beta_left, beta_left, m_left)
                    + _normal_integral(-beta_left, T(0))
                )
                * scale_left
                + (
                    _normal_integral(T(0), beta_right)
                    + _powerlaw_integral(-beta_right, beta_right, m_right)
                    - _powerlaw_integral(-z, beta_right, m_right)
                )
                * scale_right
            ) / norm
    return r


_generate_wrappers(globals())



"""
Truncated normal distribution.

See Also
--------
scipy.stats.truncnorm: Scipy equivalent.
"""

import numpy as np
from . import norm as _norm
from ._util import _jit, _generate_wrappers, _prange, _seed, _rvs_jit

_doc_par = """
xmin : float
    Lower edge of the distribution.
xmin : float
    Upper edge of the distribution.
loc : float
    Location of the mode.
scale : float
    Width parameter.
"""


@_jit(4)
def _logpdf(x, xmin, xmax, loc, scale):
    T = type(scale)
    scale2 = T(1) / scale
    z = (x - loc) * scale2
    zmin = (xmin - loc) * scale2
    zmax = (xmax - loc) * scale2
    scale *= _norm._cdf1(zmax) - _norm._cdf1(zmin)
    for i in _prange(len(z)):
        if zmin <= z[i] < zmax:
            z[i] = _norm._logpdf1(z[i]) - np.log(scale)
        else:
            z[i] = -T(np.inf)
    return z


@_jit(4)
def _pdf(x, xmin, xmax, loc, scale):
    return np.exp(_logpdf(x, xmin, xmax, loc, scale))


@_jit(4)
def _cdf(x, xmin, xmax, loc, scale):
    scale = type(scale)(1) / scale
    r = (x - loc) * scale
    zmin = (xmin - loc) * scale
    zmax = (xmax - loc) * scale
    pmin = _norm._cdf1(zmin)
    pmax = _norm._cdf1(zmax)
    for i in _prange(len(r)):
        if zmin <= r[i]:
            if r[i] < zmax:
                r[i] = (_norm._cdf1(r[i]) - pmin) / (pmax - pmin)
            else:
                r[i] = 1.0
        else:
            r[i] = 0.0
    return r


@_jit(4, cache=False)
def _ppf(p, xmin, xmax, loc, scale):
    scale2 = type(scale)(1) / scale
    zmin = (xmin - loc) * scale2
    zmax = (xmax - loc) * scale2
    pmin = _norm._cdf1(zmin)
    pmax = _norm._cdf1(zmax)
    r = p * (pmax - pmin) + pmin
    for i in _prange(len(r)):
        r[i] = _norm._ppf1(r[i])
    return scale * r + loc


@_rvs_jit(4, cache=False)
def _rvs(xmin, xmax, loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, xmin, xmax, loc, scale)


_generate_wrappers(globals())



"""
Normal distribution.

See Also
--------
scipy.stats.norm: Scipy equivalent.
"""
import numpy as np
from ._special import ndtri as _ndtri
from ._util import _jit, _trans, _generate_wrappers, _prange, _seed, _rvs_jit
from math import erf as _erf

_doc_par = """
loc : float
    Location of the mode of the distribution.
scale : float
    Standard deviation.
"""


@_jit(-1)
def _logpdf1(z):
    T = type(z)
    return -T(0.5) * (z * z + T(np.log(2 * np.pi)))


@_jit(-1)
def _cdf1(z):
    T = type(z)
    c = T(np.sqrt(0.5))
    return T(0.5) * (T(1.0) + _erf(z * c))


@_jit(-1, cache=False)  # cannot cache because of _ndtri
def _ppf1(p):
    T = type(p)
    return T(_ndtri(p))


@_jit(2)
def _logpdf(x, loc, scale):
    r = _trans(x, loc, scale)
    for i in _prange(len(r)):
        r[i] = _logpdf1(r[i]) - np.log(scale)
    return r


@_jit(2)
def _pdf(x, loc, scale):
    return np.exp(_logpdf(x, loc, scale))


@_jit(2)
def _cdf(x, loc, scale):
    r = _trans(x, loc, scale)
    for i in _prange(len(r)):
        r[i] = _cdf1(r[i])
    return r


@_jit(2, cache=False)
def _ppf(p, loc, scale):
    r = np.empty_like(p)
    for i in _prange(len(r)):
        r[i] = scale * _ppf1(p[i]) + loc
    return r


@_rvs_jit(2, cache=False)
def _rvs(loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, loc, scale)


_generate_wrappers(globals())



from scipy.integrate import quad
import numpy as np
from numba_stats import tsallis


def test_pdf():
    for m in (100, 1000):
        for t in (100, 1000):
            for n in (3, 5, 8):
                v, err = quad(lambda pt: tsallis.pdf(pt, m, t, n), 0, np.inf)
                assert abs(1 - v) < err


def test_cdf():
    for m in (100, 1000):
        for t in (100, 1000):
            for n in (3, 5, 8):
                for ptrange in ((0, 500), (500, 1000), (1000, 2000)):
                    v, err = quad(lambda pt: tsallis.pdf(pt, m, t, n), *ptrange)
                    v2 = np.diff(tsallis.cdf(ptrange, m, t, n))
                    assert abs(v2 - v) < err



"""
Crystal Ball distribution.

The Crystal Ball distribution replaces the lower tail of a normal distribution with
a power-law tail.

https://en.wikipedia.org/wiki/Crystal_Ball_function

See Also
--------
scipy.stats.crystalball: Scipy equivalent.
"""
from ._util import _jit, _trans, _generate_wrappers, _prange
import numpy as np
from math import erf as _erf

_doc_par = """
x : Array-like
    Random variate.
beta : float
    Distance from the mode in units of standard deviations where the Crystal
    Ball turns from a gaussian into a power law.
m : float
    Absolute value of the slope of the powerlaw tail. Must be large than 1.
"""


@_jit(-3)
def _log_powerlaw(z, beta, m):
    T = type(beta)
    c = -T(0.5) * beta * beta
    log_a = m * np.log(m / beta) + c
    b = m / beta - beta
    return log_a - m * np.log(b - z)


@_jit(-3)
def _powerlaw_integral(z, beta, m):
    T = type(beta)
    exp_beta = np.exp(-T(0.5) * beta * beta)
    a = (m / beta) ** m * exp_beta
    b = m / beta - beta
    m1 = m - type(m)(1)
    return a * (b - z) ** -m1 / m1


@_jit(-2)
def _normal_integral(a, b):
    T = type(a)
    sqrt_half = np.sqrt(T(0.5))
    return sqrt_half * np.sqrt(T(np.pi)) * (_erf(b * sqrt_half) - _erf(a * sqrt_half))


@_jit(-3)
def _log_density(z, beta, m):
    if z < -beta:
        return _log_powerlaw(z, beta, m)
    return -0.5 * z * z


@_jit(4)
def _logpdf(x, beta, m, loc, scale):
    z = _trans(x, loc, scale)
    norm = scale * (
        _powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, type(beta)(np.inf))
    )
    c = np.log(norm)
    for i in _prange(len(z)):
        z[i] = _log_density(z[i], beta, m) - c
    return z


@_jit(4)
def _pdf(x, beta, m, loc, scale):
    return np.exp(_logpdf(x, beta, m, loc, scale))


@_jit(4)
def _cdf(x, beta, m, loc, scale):
    z = _trans(x, loc, scale)
    norm = _powerlaw_integral(-beta, beta, m) + _normal_integral(
        -beta, type(beta)(np.inf)
    )
    for i in _prange(len(z)):
        if z[i] < -beta:
            z[i] = _powerlaw_integral(z[i], beta, m) / norm
        else:
            z[i] = (
                _powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, z[i])
            ) / norm
    return z


_generate_wrappers(globals())



import scipy.stats as sc
import numpy as np
from numba_stats import expon


def test_pdf():
    x = np.linspace(-5, 5, 20)
    got = expon.pdf(x, 1, 2)
    expected = sc.expon.pdf(x, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(-5, 5, 20) + 3
    got = expon.cdf(x, 3, 2)
    expected = sc.expon.cdf(x, 3, 2)
    np.testing.assert_allclose(got, expected)


def test_ppf():
    p = np.linspace(0, 1, 20)
    got = expon.ppf(p, 1, 2)
    expected = sc.expon.ppf(p, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_rvs():
    args = 1, 2
    x = expon.rvs(*args, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: expon.cdf(x, *args))
    assert r.pvalue > 0.01



"""
Continuous Poisson distribution.

The Poisson distribution is for discrete values, but the cdf can be generalised
to arbitrary real values. It was assumed that this distribution could be useful
to describe weighted data, but it is currently unclear how useful the
distribution is.

The pdf cannot be expressed in tabulated functions:

d G(x, mu)/d x = ln(mu) G(x, mu) + mu T(3, x, mu)

where G(x, mu) is the upper incomplete gamma function and T(m, s, x) is a
special case of the Meijer G-function, see
https://en.wikipedia.org/wiki/Incomplete_gamma_function#Derivatives

There is a Meijer G-function implemented in mpmath, but I don't know how to use it.
"""
from ._special import gammaincc as _gammaincc
from ._util import _jit, _generate_wrappers, _prange
import numpy as np

_doc_par = """
mu : float
    Expected value.
"""


@_jit(1, cache=False)
def _cdf(x, mu):
    r = np.empty_like(x)
    one = type(x[0])(1)
    for i in _prange(len(x)):
        r[i] = _gammaincc(x[i] + one, mu)
    return r


_generate_wrappers(globals())



import scipy.stats as sc
import numpy as np
import numba as nb
from numba_stats import lognorm
from numpy.testing import assert_allclose


def test_pdf():
    x = np.linspace(0, 5, 10)
    got = lognorm.pdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.pdf(x, 1.5, 0.1, 1.2)
    assert_allclose(got, expected)


def test_logpdf():
    x = np.linspace(0, 5, 10)
    got = lognorm.logpdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.logpdf(x, 1.5, 0.1, 1.2)
    assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(0, 5, 10)
    got = lognorm.cdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.cdf(x, 1.5, 0.1, 1.2)
    assert_allclose(got, expected)


def test_ppf():
    p = np.linspace(0, 1, 10)
    got = lognorm.ppf(p, 1.5, 0.1, 1.2)
    expected = sc.lognorm.ppf(p, 1.5, 0.1, 1.2)
    assert_allclose(got, expected)


def test_rvs():
    args = 1.5, 0.1, 1.2
    x = lognorm.rvs(*args, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: lognorm.cdf(x, *args))
    assert r.pvalue > 0.01


def test_njit():
    @nb.njit
    def test(x):
        a = lognorm.logpdf(x, 1.0, 0.0, 1.0)
        b = lognorm.pdf(x, 1.0, 0.0, 1.0)
        c = lognorm.cdf(x, 1.0, 0.0, 1.0)
        d = lognorm.ppf(c, 1.0, 0.0, 1.0)
        return a, b, c, d

    x = np.linspace(0, 3, 10)
    a, b, c, d = test(x)

    assert_allclose(a, lognorm.logpdf(x, 1.0, 0.0, 1.0))
    assert_allclose(b, lognorm.pdf(x, 1.0, 0.0, 1.0))
    assert_allclose(c, lognorm.cdf(x, 1.0, 0.0, 1.0))
    assert_allclose(d, x)



"""
Empirical density distribution formed by a Bernstein polynomial.

The Bernstein polynomial basis is better suited to model a probability distribution
than the Chebychev basis, since it is possible to implement the constraint
f(x; p) >= 0 with simple parameter limits p >= 0 (where p is a vector).

The density function and its integral are not normalised. Normalising would create
several issues while providing no practical benefit. Use this function in an extended
maximum-likelihood fit and do not multiply it with a scale. The scale is implicit in the
parameter vector beta. If you really need to know the scale after the fit (usually you
do not), use :func:`integral` to compute it.

See Also
--------
scipy.interpolate.BPoly: Bernstein polynomials in Scipy.
"""

import numpy as np
from ._util import _jit, _Floats, _generate_wrappers, _readonly_carray, _trans


@_jit([T[:](_readonly_carray(T), _readonly_carray(T)) for T in _Floats])
def _de_castlejau(z, beta):
    # De Casteljau algorithm, numerically stable
    n = len(beta)
    res = np.full_like(z, np.nan)
    betai = np.empty_like(beta)
    # not sure how to parallelize this, each worker thread needs its own betai
    for i in range(len(z)):
        betai[:] = beta
        azi = 1.0 - z[i]
        for j in range(1, n):
            for k in range(n - j):
                betai[k] = betai[k] * azi + betai[k + 1] * z[i]
        res[i] = betai[0]
    return res


@_jit(0)
def _beta_int(beta):
    n = len(beta)
    r = np.zeros(n + 1, dtype=beta.dtype)
    for j in range(1, n + 1):
        for k in range(j):
            r[j] += beta[k]
    r *= 1.0 / n
    return r


@_jit([T[:](_readonly_carray(T), _readonly_carray(T), T, T) for T in _Floats])
def _density(x, beta, xmin, xmax):
    """
    Return density described by a Bernstein polynomial.

    The function is non-negative, if each element of the parameter vector beta is non-
    negative, see module description.

    This function is by design equivalent to
    ``scipy.interpolate.BPoly(np.array(beta)[:, np.newaxis], [x[0], x[-1]])(x)``.

    Parameters
    ----------
    x : ArrayLike
        Values where the density is evaluated.
    beta : ArrayLike
        Vector of parameters (1D).
    xmin : float
        Lower edge of the domain of x.
    xmax : float
        Upper edge of the domain of x.

    Returns
    -------
    ndarray
        Function values.

    Examples
    --------
    >>> import numpy as np
    >>> from numba_stats import bernstein
    >>> x = np.linspace(-1, 1)
    >>> y = bernstein.density(x, [1, 2], -1, 1)

    See Also
    --------
    scipy.interpolate.BPoly
    """
    z = _trans(x, xmin, xmax - xmin)
    return _de_castlejau(z, beta)


@_jit(
    [T[:](_readonly_carray(T), _readonly_carray(T), T, T) for T in _Floats], cache=True
)
def _integral(x, beta, xmin, xmax):
    """
    Return integral of a Bernstein polynomial from xmin to x.

    Parameters
    ----------
    x : ArrayLike
        Values up to which the integral is computed, starting from xmin.
    beta : ArrayLike
        Vector of parameters (1D).
    xmin : float
        Lower edge of the domain of x.
    xmax : float
        Upper edge of the domain of x.

    Returns
    -------
    ndarray
        Integral values.

    Examples
    --------
    >>> import numpy as np
    >>> from numba_stats import bernstein
    >>> x = np.linspace(-1, 1)
    >>> y = bernstein.integral(x, [1, 2], -1, 1)

    See Also
    --------
    scipy.interpolate.BPoly
    """
    scale = xmax - xmin
    z = _trans(x, xmin, scale)
    beta = _beta_int(beta) * scale
    return _de_castlejau(z, beta)


def _wrap(fn):
    def outer(x, beta, xmin, xmax):
        shape = np.shape(x)
        args = []
        for arg in (x, beta):
            arg = np.array(arg).flatten()
            if arg.dtype.kind != "f":
                arg = arg.astype(float)
            args.append(arg)
        return fn(*args, xmin, xmax).reshape(shape)

    return outer


def _type_check(x, beta, xmin, xmax):
    from numba.types import Array
    from numba.core.errors import TypingError

    for arg in (x, beta):
        if not (isinstance(arg, Array) and arg.dtype in _Floats):
            raise TypingError(
                "first two arguments must be arrays of floating point type"
            )
    T = type(arg.dtype)
    for i, tp in enumerate((xmin, xmax)):
        if not isinstance(tp, T):
            raise TypingError(f"argument {i+1} must be of type {tp}")


_generate_wrappers(globals())


def __getattr__(key):
    # Temporary hack to maintain backward compatibility
    import warnings
    from numpy import VisibleDeprecationWarning

    if key in ("scaled_pdf", "scaled_cdf"):
        r = {"scaled_pdf": "density", "scaled_cdf": "integral"}
        warnings.warn(
            f"bernstein.{key} is deprecated and will be removed in a future release, "
            f"use bernstein.{r[key]} instead",
            VisibleDeprecationWarning,
            1,
        )
        return globals()[r[key]]
    raise AttributeError



from numba_stats import uniform
import numpy as np
import scipy.stats as sc
from numpy.testing import assert_allclose


def test_pdf():
    x = np.linspace(-1.1, 1.1, 10)
    got = uniform.pdf(x, -1, 2)
    expected = sc.uniform.pdf(x, -1, 2)
    assert_allclose(got, expected)

    got = uniform.pdf(1, -1, 3)
    expected = sc.uniform.pdf(1, -1, 3)
    assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(-1.1, 1.1, 10)
    got = uniform.cdf(x, -1, 2)
    expected = sc.uniform.cdf(x, -1, 2)
    assert_allclose(got, expected)

    got = uniform.cdf(1, -1, 3)
    expected = sc.uniform.cdf(1, -1, 3)
    assert_allclose(got, expected)


def test_ppf():
    x = np.linspace(0, 1, 10)
    got = uniform.ppf(x, -1, 2)
    expected = sc.uniform.ppf(x, -1, 2)
    assert_allclose(got, expected)

    got = uniform.ppf(0.5, -1, 3)
    expected = sc.uniform.ppf(0.5, -1, 3)
    assert_allclose(got, expected)


def test_rvs():
    args = -1, 2
    x = uniform.rvs(*args, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: uniform.cdf(x, *args))
    assert r.pvalue > 0.01



import numpy as np
from numba_stats import poisson
import scipy.stats as sc
import pytest


@pytest.mark.parametrize("mu", np.linspace(0, 3, 5))
def test_pmf(mu):
    k = np.arange(10)
    got = poisson.pmf(k, mu)
    expected = sc.poisson.pmf(k, mu)
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("mu", np.linspace(0, 3, 5))
def test_cdf(mu):
    k = np.arange(10)
    got = poisson.cdf(k, mu)
    expected = sc.poisson.cdf(k, mu)
    np.testing.assert_allclose(got, expected)



"""
Lognormal distribution.

See Also
--------
scipy.stats.lognorm: Scipy equivalent.
"""
import numpy as np
from . import norm as _norm
from ._util import _jit, _trans, _generate_wrappers, _prange, _seed, _rvs_jit

_doc_par = """
s : float
    Standard deviation of the corresponding normal distribution of exp(x).
loc : float
    Shift of the distribution.
scale : float
    Equal to exp(mu) of the corresponding normal distribution of exp(x).
"""


@_jit(3)
def _logpdf(x, s, loc, scale):
    r = _trans(x, loc, scale)
    for i in _prange(len(r)):
        if r[i] > 0:
            r[i] = -0.5 * np.log(r[i]) ** 2 / s**2 - np.log(
                s * r[i] * np.sqrt(2 * np.pi) * scale
            )
        else:
            r[i] = -np.inf
    return r


@_jit(3)
def _pdf(x, s, loc, scale):
    return np.exp(_logpdf(x, s, loc, scale))


@_jit(3)
def _cdf(x, s, loc, scale):
    r = _trans(x, loc, scale)
    for i in _prange(len(r)):
        if r[i] <= 0:
            r[i] = 0.0
        else:
            z = np.log(r[i]) / s
            r[i] = _norm._cdf1(z)
    return r


@_jit(3, cache=False)  # no cache because of norm._ppf
def _ppf(p, s, loc, scale):
    r = np.empty_like(p)
    for i in _prange(len(p)):
        r[i] = np.exp(s * _norm._ppf1(p[i]))
    return scale * r + loc


@_rvs_jit(3, cache=False)
def _rvs(s, loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, s, loc, scale)


_generate_wrappers(globals())



"""
Tsallis-Hagedorn distribution.

A generalisation (q-analog) of the exponential distribution based on Tsallis entropy. It
approximately describes the pT distribution charged particles produced in high-energy
minimum bias particle collisions.
"""

import numpy as np
from ._util import _jit, _generate_wrappers

_doc_par = """
m : float
    Mass of the particle.
t :  float
    Width parameter.
n : float
    Absolute value of the exponent of the power law.
"""


@_jit(3)
def _pdf(x, m, t, n):
    # Formula from CMS, Eur. Phys. J. C (2012) 72:2164
    if n <= 2:
        raise ValueError("n > 2 is required")

    T = type(m)
    mt = np.sqrt(m * m + x * x)
    nt = n * t
    c = (n - T(1)) * (n - T(2)) / (nt * (nt + (n - T(2)) * m))
    return c * x * (T(1) + (mt - m) / nt) ** -n


@_jit(3)
def _cdf(x, m, t, n):
    # Formula computed from tsallis_pdf with Sympy, then simplified by hand
    if n <= 2:
        raise ValueError("n > 2 is required")

    T = type(m)
    mt = np.sqrt(m * m + x * x)
    nt = n * t
    return (
        ((mt - m) / nt + T(1)) ** (T(1) - n)
        * (m + mt - n * (mt + t))
        / (m * (n - T(2)) + nt)
    )


_generate_wrappers(globals())



import numpy as np
from numba_stats import truncnorm
import scipy.stats as sc
from scipy.integrate import quad
import pytest


@pytest.mark.parametrize("mu", (0, -1, 1))
@pytest.mark.parametrize("sigma", (1, 0.5, 2))
def test_truncnorm(mu, sigma):
    got = quad(lambda x: truncnorm.pdf(x, -1, 1, mu, sigma), -10, 10)[0]
    expected = 1.0
    np.testing.assert_allclose(got, expected)


def test_logpdf():
    x = np.linspace(-1, 5, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    got = truncnorm.logpdf(x, xmin, xmax, mu, sigma)
    z = (x - mu) / sigma
    zmin = (xmin - mu) / sigma
    zmax = (xmax - mu) / sigma
    expected = sc.truncnorm.logpdf(z, zmin, zmax) - np.log(sigma)
    np.testing.assert_allclose(got, expected)


def test_pdf():
    x = np.linspace(-1, 5, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    got = truncnorm.pdf(x, xmin, xmax, mu, sigma)
    z = (x - mu) / sigma
    zmin = (xmin - mu) / sigma
    zmax = (xmax - mu) / sigma
    expected = sc.truncnorm.pdf(z, zmin, zmax) / sigma
    np.testing.assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(-1, 5, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    got = truncnorm.cdf(x, xmin, xmax, mu, sigma)
    z = (x - mu) / sigma
    zmin = (xmin - mu) / sigma
    zmax = (xmax - mu) / sigma
    expected = sc.truncnorm.cdf(z, zmin, zmax)
    np.testing.assert_allclose(got, expected)


def test_ppf():
    expected = np.linspace(0, 1, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    x = truncnorm.ppf(expected, mu, sigma, xmin, xmax)
    got = truncnorm.cdf(x, mu, sigma, xmin, xmax)
    np.testing.assert_allclose(got, expected, atol=1e-14)


def test_rvs():
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    x = truncnorm.rvs(xmin, xmax, mu, sigma, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: truncnorm.cdf(x, xmin, xmax, mu, sigma))
    assert r.pvalue > 0.01



from numba_stats import cruijff
import numpy as np
from scipy.stats import norm
from numpy.testing import assert_allclose


def test_density_1():
    x = np.linspace(-5, 5, 100)
    got = cruijff.density(x, 0, 0, 0, 1, 1) / np.sqrt(2 * np.pi)
    expected = norm().pdf(x)
    assert_allclose(got, expected)


def test_density_2():
    x = np.linspace(-5, 5, 100)
    got = cruijff.density(x, 0.1, 0.2, 1.5, 2.1, 2.1)
    expected = cruijff.density((x - 1.5) / 2.1, 0.1, 0.2, 0, 1, 1)
    assert_allclose(got, expected)


def test_density_3():
    x = np.linspace(-5, 5, 100)
    y = cruijff.density(x, 0.2, 0.3, 0, 1.1, 1.2)
    dy = np.diff(y, 1)
    # density should be smooth
    assert_allclose(dy, 0, atol=0.05)
    # function rises up to x = 0
    assert np.all(dy[x[1:] < 0] > 0)
    # function falls after x = 0
    assert np.all(dy[x[1:] > 0.1] < 0)



"""
Uniform distribution.

See Also
--------
scipy.stats.uniform: Equivalent in Scipy.
"""
from ._util import _jit, _generate_wrappers, _prange, _rvs_jit, _seed
import numpy as np

_doc_par = """
a : float
    Lower edge of the distribution.
w : float
    Width of the distribution.
"""


@_jit(2)
def _logpdf(x, a, w):
    r = np.empty_like(x)
    for i in _prange(len(x)):
        if a <= x[i] <= a + w:
            r[i] = -np.log(w)
        else:
            r[i] = -np.inf
    return r


@_jit(2)
def _pdf(x, a, w):
    return np.exp(_logpdf(x, a, w))


@_jit(2)
def _cdf(x, a, w):
    r = np.empty_like(x)
    for i in _prange(len(x)):
        if a <= x[i]:
            if x[i] <= a + w:
                r[i] = (x[i] - a) / w
            else:
                r[i] = 1
        else:
            r[i] = 0
    return r


@_jit(2)
def _ppf(p, a, w):
    return w * p + a


@_rvs_jit(2)
def _rvs(a, w, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, a, w)


_generate_wrappers(globals())



"""
Laplace distribution.

See Also
--------
scipy.stats.laplace: Scipy equivalent.
"""
import numpy as np
from ._util import _jit, _trans, _generate_wrappers, _prange, _rvs_jit, _seed

_doc_par = """
loc : float
    Location of the mode.
scale : float
    Standard deviation.
"""


@_jit(-1)
def _cdf1(z):
    return 1.0 - 0.5 * np.exp(-z) if z > 0 else 0.5 * np.exp(z)


@_jit(-1)
def _ppf1(p):
    return -np.log(2 * (1 - p)) if p > 0.5 else np.log(2 * p)


@_jit(2)
def _logpdf(x, loc, scale):
    z = _trans(x, loc, scale)
    r = np.empty_like(z)
    for i in _prange(len(r)):
        r[i] = np.log(0.25) - np.abs(z[i])
    return r


@_jit(2)
def _pdf(x, loc, scale):
    return np.exp(_logpdf(x, loc, scale))


@_jit(2)
def _cdf(x, loc, scale):
    z = _trans(x, loc, scale)
    for i in _prange(len(z)):
        z[i] = _cdf1(z[i])
    return z


@_jit(2)
def _ppf(p, loc, scale):
    z = np.empty_like(p)
    for i in _prange(len(z)):
        z[i] = _ppf1(p[i])
    return scale * z + loc


@_rvs_jit(2)
def _rvs(loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, loc, scale)


_generate_wrappers(globals())



"""
Voigtian distribution.

This is the convolution of a Cauchy distribution with a normal distribution.

There is a closed form for the cdf, but the required hypergeometric function is not
implemented anywhere.

https://en.wikipedia.org/wiki/Voigt_profile

See Also
--------
scipy.special.voigt_profile: Equvialent in Scipy.
"""
from ._special import voigt_profile as _voigt
from ._util import _jit, _generate_wrappers, _prange
import numpy as np

_doc_par = """
gamma : float
    The half-width at half-maximum of the Cauchy distribution part.
loc : float
    Location of the mode.
scale : float
    Standard deviation of the normal distribution.
"""


@_jit(3, cache=False)
def _pdf(x, gamma, loc, scale):
    r = np.empty_like(x)
    for i in _prange(len(x)):
        r[i] = _voigt(x[i] - loc, scale, gamma)
    return r


_generate_wrappers(globals())



"""
Student's t distribution.

See Also
--------
scipy.stats.t: Scipy equivalent.
"""
import numpy as np
from ._special import stdtr as _stdtr, stdtrit as _stdtrit
from ._util import _jit, _trans, _generate_wrappers, _prange, _seed, _rvs_jit
from math import lgamma as _lgamma

_doc_par = """
df : float
    Degrees of freedom.
loc : float
    Location of the mode.
scale : float
    Width parameter.
"""


@_jit(3, cache=False)
def _logpdf(x, df, loc, scale):
    T = type(df)
    z = _trans(x, loc, scale)
    k = T(0.5) * (df + T(1))
    c = _lgamma(k) - _lgamma(T(0.5) * df)
    c -= T(0.5) * np.log(df * T(np.pi))
    c -= np.log(scale)
    for i in _prange(len(z)):
        z[i] = -k * np.log(T(1) + (z[i] * z[i]) / df) + c
    return z


@_jit(3, cache=False)
def _pdf(x, df, loc, scale):
    return np.exp(_logpdf(x, df, loc, scale))


@_jit(3, cache=False)
def _cdf(x, df, loc, scale):
    z = _trans(x, loc, scale)
    for i in _prange(len(z)):
        z[i] = _stdtr(df, z[i])
    return z


@_jit(3, cache=False)
def _ppf(p, df, loc, scale):
    T = type(df)
    r = np.empty_like(p)
    for i in _prange(len(p)):
        if p[i] == 0:
            r[i] = -T(np.inf)
        elif p[i] == 1:
            r[i] = T(np.inf)
        else:
            r[i] = _stdtrit(df, p[i])
    return scale * r + loc


@_rvs_jit(3, cache=False)
def _rvs(df, loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, df, loc, scale)


_generate_wrappers(globals())



"""
Truncated exponential distribution.

See Also
--------
scipy.stats.truncexpon: Scipy equivalent.
"""
import numpy as np
from ._util import _jit, _trans, _generate_wrappers, _prange, _rvs_jit, _seed
from . import expon as _expon

_doc_par = """
xmin : float
    Lower edge of the distribution.
xmax : float
    Upper edge of the distribution.
loc : float
    Location of the mode.
scale : float
    Width parameter.
"""


@_jit(4)
def _logpdf(x, xmin, xmax, loc, scale):
    T = type(xmin)
    z = _trans(x, loc, scale)
    scale2 = T(1) / scale
    zmin = (xmin - loc) * scale2
    zmax = (xmax - loc) * scale2
    c = np.log(scale * (_expon._cdf1(zmax) - _expon._cdf1(zmin)))
    zmin = max(zmin, T(0))
    for i in _prange(len(z)):
        if zmin <= z[i] < zmax:
            z[i] = -z[i] - c
        else:
            z[i] = -np.inf
    return z


@_jit(4)
def _pdf(x, xmin, xmax, loc, scale):
    return np.exp(_logpdf(x, xmin, xmax, loc, scale))


@_jit(4)
def _cdf(x, xmin, xmax, loc, scale):
    T = type(xmin)
    z = _trans(x, loc, scale)
    scale2 = T(1) / scale
    zmin = (xmin - loc) * scale2
    zmax = (xmax - loc) * scale2
    pmin = _expon._cdf1(zmin)
    pmax = _expon._cdf1(zmax)
    scale3 = T(1) / (pmax - pmin)
    zmin = max(zmin, T(0))
    for i in _prange(len(z)):
        if zmin <= z[i]:
            if z[i] < zmax:
                z[i] = (_expon._cdf1(z[i]) - pmin) * scale3
            else:
                z[i] = 1
        else:
            z[i] = 0
    return z


@_jit(4)
def _ppf(p, xmin, xmax, loc, scale):
    zmin = (xmin - loc) / scale
    zmax = (xmax - loc) / scale
    pmin = _expon._cdf1(zmin)
    pmax = _expon._cdf1(zmax)
    z = p * (pmax - pmin) + pmin
    for i in _prange(len(z)):
        z[i] = _expon._ppf1(z[i])
    return z * scale + loc


@_rvs_jit(4)
def _rvs(xmin, xmax, loc, scale, size, random_state):
    _seed(random_state)
    p = np.random.uniform(0, 1, size)
    return _ppf(p, xmin, xmax, loc, scale)


_generate_wrappers(globals())



import pytest
import pkgutil
import numba_stats
import subprocess as subp
import importlib

all_modules = []
for _, modname, ispkg in pkgutil.walk_packages(numba_stats.__path__):
    if modname.startswith("_"):
        continue
    all_modules.append(modname)


@pytest.mark.parametrize("module", all_modules)
def test_all(module):
    pytest.importorskip("pydocstyle")
    m = importlib.import_module(f"numba_stats.{module}")
    r = subp.run(["python", "-m", "pydocstyle", m.__file__], stdout=subp.PIPE)
    rc = int(r.returncode)
    assert rc == 0, r.stdout.decode("utf8")



import scipy.stats as sc
import numpy as np
import numba as nb
from numba_stats import norm
from numpy.testing import assert_allclose
import pytest


def test_pdf_one():
    x = 1
    got = norm.pdf(x, 1, 2)
    expected = sc.norm.pdf(x, 1, 2)
    assert_allclose(got, expected)


def test_pdf():
    x = np.linspace(-5, 5, 10)
    got = norm.pdf(x, 1, 2)
    expected = sc.norm.pdf(x, 1, 2)
    assert_allclose(got, expected)


def test_logpdf():
    x = np.linspace(-5, 5, 10)
    got = norm.logpdf(x, 1, 2)
    expected = sc.norm.logpdf(x, 1, 2)
    assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(-5, 5, 10)
    got = norm.cdf(x, 1, 2)
    expected = sc.norm.cdf(x, 1, 2)
    assert_allclose(got, expected)


def test_ppf():
    p = np.linspace(0, 1, 10)
    got = norm.ppf(p, 0, 1)
    expected = sc.norm.ppf(p)
    assert_allclose(got, expected)

    got = norm.ppf(0.5, 0, 1)
    expected = sc.norm.ppf(0.5, 0, 1)
    assert_allclose(got, expected)


def test_rvs():
    mu = 2
    sigma = 3
    x = norm.rvs(mu, sigma, size=100_000, random_state=1)
    r = sc.kstest(x, lambda x: norm.cdf(x, mu, sigma))
    assert r.pvalue > 0.01


@pytest.mark.filterwarnings("error")
@pytest.mark.parametrize("fn", [norm.logpdf, norm.pdf, norm.cdf, norm.ppf])
@pytest.mark.parametrize("parallel", [False, True])
def test_njit(fn, parallel):
    @nb.njit(parallel=parallel, fastmath=True)
    def test(x):
        return fn(x, 0.0, 1.0)

    x = np.linspace(-3, 3, 1000)
    y = test(x)

    assert_allclose(y, fn(x, 0, 1))



from numba_stats import crystalball as cb
import numpy as np
import pytest
from scipy import stats as sc
from scipy.integrate import quad
from numpy.testing import assert_allclose


@pytest.mark.parametrize("beta", (5, 1, 0.1))
@pytest.mark.parametrize("m", (1.1, 2, 3))
def test_powerlaw_integral(beta, m):
    expected = quad(lambda z: np.exp(cb._log_powerlaw(z, beta, m)), -np.inf, -beta)[0]
    got = cb._powerlaw_integral(-beta, beta, m)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (5, 1, 0.1))
@pytest.mark.parametrize("z", (0, 1, 10, np.inf))
def test_normal_integral(beta, z):
    expected = quad(lambda z: np.exp(-0.5 * z**2), -beta, z)[0]
    got = cb._normal_integral(-beta, z)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (5, 1, 0.1))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_logpdf(beta, m):
    scale = 1.5
    x = np.linspace(-10, 5, 10)
    got = cb.logpdf(x, beta, m, 0, scale)
    expected = sc.crystalball.logpdf(x, beta, m, 0, scale)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (5, 1, 0.1))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_pdf(beta, m):
    scale = 1.5
    x = np.linspace(-10, 5, 10)
    got = cb.pdf(x, beta, m, 0, scale)
    expected = sc.crystalball.pdf(x, beta, m, 0, scale)
    assert_allclose(got, expected)


@pytest.mark.parametrize("beta", (5, 1, 0.1))
@pytest.mark.parametrize("m", (1.001, 2, 3))
def test_cdf(beta, m):
    scale = 1.5
    x = np.linspace(-10, 5, 10)
    got = cb.cdf(x, beta, m, 0, scale)
    expected = sc.crystalball.cdf(x, beta, m, 0, scale)
    assert_allclose(got, expected)



from numba_stats import poisson, cpoisson
import numpy as np
import pytest


@pytest.mark.parametrize("mu", (0.1, 0.5, 1.0, 2.0))
def test_cdf(mu):
    k = np.arange(10)
    got = cpoisson.cdf(k, mu)
    expected = poisson.cdf(k, mu)
    np.testing.assert_allclose(got, expected)
